from typing import Dict, Any
from jinja2 import Environment, BaseLoader

from unipipeline.utils.camel_case import camel_case


def template(definition: str, data: Dict[str, Any]) -> str:
    assert isinstance(definition, str), f"definition must be str. {type(definition)} given"

    env = Environment(loader=BaseLoader())
    env.filters['camel'] = camel_case
    rtemplate = env.from_string(definition)
    result = rtemplate.render(**data)

    return result



from unipipeline import UniWorker

from example.messages.input_message import InputMessage


class InputWorker(UniWorker):
    def handle_message(self, message: InputMessage) -> None:
        print("!!!")
        raise NotImplementedError('method handle_message must be specified for class "InputWorker"')



import json
import logging
from logging import Logger
from typing import Callable
from uuid import uuid4

from unipipeline import UniBroker, UniMessageMeta, UniBrokerMessageManager


class UniLogBroker(UniBroker):
    def mk_logger(self) -> Logger:
        return logging.getLogger(__name__)

    def mk_log_prefix(self) -> str:
        return f'{type(self).__name__} {self._definition.name}::{uuid4()} :'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._logger = self.mk_logger()
        self._logging_prefix = self.mk_log_prefix()

    def connect(self) -> None:
        self._logger.info(f'{self._logging_prefix} connect')

    def close(self) -> None:
        self._logger.info(f'{self._logging_prefix} close')

    def consume(self, topic: str, processor: Callable[[UniMessageMeta, UniBrokerMessageManager], None], consumer_tag: str, worker_name: str, prefetch: int = 1) -> None:
        self._logger.info(f'{self._logging_prefix} consume {consumer_tag} :: {worker_name}')

    def publish(self, topic: str, meta: UniMessageMeta) -> None:
        self._logger.info(f'{self._logging_prefix} publish {json.dumps(meta.dict())}')



from typing import Dict, Type, TypeVar, Any

from unipipeline.modules.uni_worker import UniWorker
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_broker import UniBroker
from unipipeline.modules.uni_config import UniConfig
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition

TMessage = TypeVar('TMessage')
T = TypeVar('T')


class UniMediator:
    def __init__(
        self,
        config: UniConfig,
    ) -> None:
        self._config = config
        self._connected_brokers: Dict[str, UniBroker] = dict()
        self._worker_definition_by_type: Dict[Any, UniWorkerDefinition] = dict()
        self._worker_instance_indexes: Dict[str, UniWorker] = dict()

    def get_worker(self, name: str, singleton: bool = True) -> UniWorker[UniMessage]:
        if not singleton:
            w_def = self._config.workers[name]
            worker_type = w_def.type.import_class(UniWorker)
            return worker_type(definition=w_def, mediator=self)
        if name not in self._worker_instance_indexes:
            self._worker_instance_indexes[name] = self.get_worker(name, singleton=False)
        return self._worker_instance_indexes[name]

    @property
    def config(self) -> UniConfig:
        return self._config

    def get_connected_broker_instance(self, name: str) -> UniBroker:
        if name not in self._connected_brokers:
            self._connected_brokers[name] = UniBroker.waiting_for_connection(self.config.brokers[name])
        return self._connected_brokers[name]

    def wait_related_brokers(self, worker_name: str) -> None:
        broker_names = self.config.workers[worker_name].get_related_broker_names(self.config.workers)
        for bn in broker_names:
            self.get_connected_broker_instance(bn)

    def load_workers(self, uni_type: Type[T]) -> None:
        if self._worker_definition_by_type:
            return
        for wd in self.config.workers.values():
            self._worker_definition_by_type[wd.type.import_class(uni_type)] = wd

    def get_worker_definition_by_type(self, worker_type: Any, uni_type: Type[T]) -> UniWorkerDefinition:
        self.load_workers(uni_type)
        return self._worker_definition_by_type[worker_type]




from typing import Callable

from unipipeline import UniBroker, UniMessageMeta, UniBrokerMessageManager


class LogBroker(UniBroker):
    def connect(self) -> None:
        raise NotImplementedError('method connect must be specified for class "LogBroker"')

    def close(self) -> None:
        raise NotImplementedError('method close must be specified for class "LogBroker"')

    def consume(self, topic: str, processor: Callable[[UniMessageMeta, UniBrokerMessageManager], None], consumer_tag: str, worker_name: str, prefetch: int = 1) -> None:
        raise NotImplementedError('method consume must be specified for class "LogBroker"')

    def publish(self, topic: str, meta: UniMessageMeta) -> None:
        raise NotImplementedError('method publish must be specified for class "LogBroker"')



from typing import Dict, Any, List

import yaml  # type: ignore

from unipipeline.modules.uni_worker_definition import UniWorkerDefinition
from unipipeline.modules.uni_message_type_definition import UniMessageTypeDefinition
from unipipeline.modules.uni_waiting_definition import UniWaitingDefinition
from unipipeline.modules.uni_broker_definition import  UniBrokerDefinition
from unipipeline.modules.uni_broker_definition import UniMessageCodec
from unipipeline.modules.uni_broker_definition import UniBrokerKafkaPropsDefinition
from unipipeline.modules.uni_broker_definition import UniBrokerRMQPropsDefinition
from unipipeline.modules.uni_cron_task_definition import UniCronTaskDefinition
from unipipeline.utils.parse_definition import parse_definition
from unipipeline.utils.parse_type import parse_type
from unipipeline.utils.template import template

UNI_CRON_MESSAGE = "uni_cron_message"


class ConfigError(Exception):
    pass


class UniConfig:
    def __init__(self, file_path: str) -> None:
        self._file_path = file_path

        self._config: Dict[str, Any] = dict()
        self._parsed = False
        self._config_loaded = False
        self._waitings_index: Dict[str, UniWaitingDefinition] = dict()
        self._brokers_index: Dict[str, UniBrokerDefinition] = dict()
        self._messages_index: Dict[str, UniMessageTypeDefinition] = dict()
        self._workers_index: Dict[str, UniWorkerDefinition] = dict()
        self._cron_tasks_index: Dict[str, UniCronTaskDefinition] = dict()

    @property
    def brokers(self) -> Dict[str, UniBrokerDefinition[Any]]:
        self._parse()
        return self._brokers_index

    @property
    def cron_tasks(self) -> Dict[str, UniCronTaskDefinition]:
        self._parse()
        return self._cron_tasks_index

    @property
    def workers(self) -> Dict[str, UniWorkerDefinition]:
        self._parse()
        return self._workers_index

    @property
    def waitings(self) -> Dict[str, UniWaitingDefinition]:
        self._parse()
        return self._waitings_index

    @property
    def messages(self) -> Dict[str, UniMessageTypeDefinition]:
        self._parse()
        return self._messages_index

    def _load_config(self) -> Dict[str, Any]:
        if self._config_loaded:
            return self._config
        self._config_loaded = True
        with open(self._file_path, "rt") as f:
            self._config = yaml.safe_load(f)
        if not isinstance(self._config, dict):
            raise ConfigError('config must be dict')
        return self._config

    def _parse(self) -> None:
        if self._parsed:
            return

        self._waitings_index = self._parse_waitings(self._load_config())
        self._brokers_index = self._parse_brokers(self._load_config())
        self._messages_index = self._parse_messages(self._load_config())
        self._workers_index = self._parse_workers(self._load_config(), self._brokers_index, self._messages_index, self._waitings_index)
        self._cron_tasks_index = self._parse_cron_tasks(self._load_config(), self._workers_index)

        self._parsed = True

    def _parse_cron_tasks(self, config: Dict[str, Any], workers: Dict[str, UniWorkerDefinition]) -> Dict[str, UniCronTaskDefinition]:
        result = dict()
        for name, definition in parse_definition("cron", config.get("cron", dict()), dict(), {"when", "worker"}):
            w = workers[definition["worker"]]
            if w.input_message.name != UNI_CRON_MESSAGE:
                raise ValueError(f"input_message of worker '{w.name}' must be '{UNI_CRON_MESSAGE}'. '{w.input_message.name}' was given")
            result[name] = UniCronTaskDefinition(
                name=name,
                worker=w,
                when=definition["when"],
            )
        return result

    def _parse_messages(self, config: Dict[str, Any]) -> Dict[str, UniMessageTypeDefinition]:
        from unipipeline import UniModuleDefinition

        result = {
            UNI_CRON_MESSAGE: UniMessageTypeDefinition(
                name=UNI_CRON_MESSAGE,
                type=UniModuleDefinition(
                    module="unipipeline.messages.uni_cron_message",
                    class_name="UniCronMessage",
                ),
            )
        }

        for name, definition in parse_definition("messages", config["messages"], dict(), {"import_template", }):
            import_template = definition.pop("import_template")
            result[name] = UniMessageTypeDefinition(
                **definition,
                type=parse_type(template(import_template, definition)),
            )

        return result

    def _parse_waitings(self, config: Dict[str, Any]) -> Dict[str, UniWaitingDefinition]:
        result = dict()
        defaults = dict(
            retry_max_count=3,
            retry_delay_s=10,
        )
        for name, definition in parse_definition('waitings', config['waitings'], defaults, {"import_template", }):
            result[name] = UniWaitingDefinition(
                **definition,
                type=parse_type(template(definition["import_template"], definition)),
            )

        return result

    def _parse_brokers(self, config: Dict[str, Any]) -> Dict[str, UniBrokerDefinition[Any]]:
        result: Dict[str, UniBrokerDefinition] = dict()
        defaults = dict(
            retry_max_count=3,
            retry_delay_s=10,

            content_type="application/json",
            compression=None,

            exchange_name="communication",
            exchange_type="direct",

            heartbeat=600,
            blocked_connection_timeout=300,
            socket_timeout=300,
            stack_timeout=300,
            passive=False,
            durable=True,
            auto_delete=False,
            is_persistent=True,
            api_version=[0, 10],
        )
        for name, definition in parse_definition("brokers", config["brokers"], defaults, {"import_template", }):
            result[name] = UniBrokerDefinition(
                **definition,
                type=parse_type(template(definition["import_template"], definition)),
                message_codec=UniMessageCodec(
                    content_type=definition["content_type"],
                    compression=definition["compression"],
                ),
                rmq_definition=UniBrokerRMQPropsDefinition(
                    exchange_name=definition['exchange_name'],
                    heartbeat=definition['heartbeat'],
                    blocked_connection_timeout=definition['blocked_connection_timeout'],
                    socket_timeout=definition['socket_timeout'],
                    stack_timeout=definition['stack_timeout'],
                    exchange_type=definition['exchange_type'],
                ),
                kafka_definition=UniBrokerKafkaPropsDefinition(
                    api_version=definition['api_version'],
                )
            )
        return result

    def _parse_workers(
        self,
        config: Dict[str, Any],
        brokers: Dict[str, UniBrokerDefinition],
        messages: Dict[str, UniMessageTypeDefinition],
        waitings: Dict[str, UniWaitingDefinition]
    ) -> Dict[str, UniWorkerDefinition]:
        result = dict()

        out_workers = set()

        defaults = dict(
            topic="{{name}}__{{input_message.name}}",
            broker="default_broker",
            prefetch=1,
            retry_max_count=3,
            retry_delay_s=1,
            max_ttl_s=None,
            is_permanent=True,
            ack_after_success=True,
            waiting_for=[],
            output_workers=[],
        )

        for name, definition in parse_definition("workers", config["workers"], defaults, {"import_template", "input_message", "broker"}):
            for ow in definition["output_workers"]:
                out_workers.add(ow)

            br = definition["broker"]
            if br not in brokers:
                raise ConfigError(f'definition workers->{name} has invalid broker: {br}')
            definition["broker"] = brokers[br]

            im = definition["input_message"]
            if im not in messages:
                raise ConfigError(f'definition workers->{name} has invalid input_message: {im}')
            definition["input_message"] = messages[im]

            waitings_: List[UniWaitingDefinition] = list()
            for w in definition["waiting_for"]:
                if w not in waitings:
                    raise ConfigError(f'definition workers->{name} has invalid waiting_for: {w}')
                waitings_.append(waitings[w])

            definition.update(
                type=parse_type(template(definition["import_template"], definition)),
                topic=template(definition["topic"], definition),
                waitings=waitings_,
            )

            defn = UniWorkerDefinition(**definition)

            result[name] = defn

        out_intersection_workers = set(result.keys()).intersection(out_workers)
        if len(out_intersection_workers) != len(out_workers):
            raise ConfigError(f'workers definition has invalid worker_names (in output_workers prop): {", ".join(out_intersection_workers)}')

        return result



from typing import NamedTuple, List, Tuple, Optional

from crontab import CronTab  # type: ignore

from unipipeline.messages.uni_cron_message import UniCronMessage
from unipipeline.modules.uni_cron_task_definition import UniCronTaskDefinition
from unipipeline.modules.uni_worker import UniWorker
from unipipeline.modules.uni_message import UniMessage


class UniCronJob(NamedTuple):
    id: int
    name: str
    crontab: CronTab
    worker: UniWorker[UniMessage]
    message: UniCronMessage

    @staticmethod
    def new(id: int, task_def: UniCronTaskDefinition, worker: UniWorker[UniMessage]) -> 'UniCronJob':
        return UniCronJob(
            id=id,
            name=task_def.name,
            crontab=CronTab(task_def.when),
            worker=worker,
            message=UniCronMessage(
                task_name=task_def.name
            )
        )

    @staticmethod
    def search_next_tasks(all_tasks: List['UniCronJob']) -> Tuple[Optional[int], List['UniCronJob']]:
        min_delay: Optional[int] = None
        notification_list: List[UniCronJob] = []
        for cj in all_tasks:
            sec = int(cj.crontab.next(default_utc=False))
            if min_delay is None:
                min_delay = sec
            if sec < min_delay:
                notification_list.clear()
                min_delay = sec
            if sec <= min_delay:
                notification_list.append(cj)

        return min_delay, notification_list

    def send(self) -> None:
        self.worker.send(self.message)



from unipipeline.brokers.uni_amqp_broker import AmqpUniBroker, AmqpUniBrokerMessageManager, RMQConnectionObj
from unipipeline.brokers.uni_kafka_broker import UniKafkaBroker, KafkaUniBrokerMessageManager
from unipipeline.modules.uni import Uni
from unipipeline.modules.uni_broker import UniBrokerMessageManager, UniBroker
from unipipeline.modules.uni_broker_definition import UniMessageCodec, UniBrokerKafkaPropsDefinition, UniBrokerRMQPropsDefinition, UniBrokerDefinition
from unipipeline.modules.uni_mediator import UniMediator
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta
from unipipeline.modules.uni_message_type_definition import UniMessageTypeDefinition
from unipipeline.modules.uni_module_definition import UniModuleDefinition
from unipipeline.modules.uni_waiting_definition import UniWaitingDefinition
from unipipeline.modules.uni_wating import UniWaiting
from unipipeline.modules.uni_worker import UniWorker
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition
from unipipeline.utils.connection_pool import ConnectionObj, ConnectionRC, ConnectionManager, ConnectionPool, connection_pool
from unipipeline.utils.serializer_registry import SerializersRegistry, serializer_registry, compressor_registry

__all__ = (
    "AmqpUniBroker",
    "AmqpUniBrokerMessageManager",
    "RMQConnectionObj",
    "UniKafkaBroker",
    "KafkaUniBrokerMessageManager",
    "Uni",
    "UniBrokerMessageManager",
    "UniBroker",
    "UniMessageCodec",
    "UniBrokerKafkaPropsDefinition",
    "UniBrokerRMQPropsDefinition",
    "UniBrokerDefinition",
    "UniMediator",
    "UniMessage",
    "UniMessageMeta",
    "UniMessageTypeDefinition",
    "UniModuleDefinition",
    "UniWaitingDefinition",
    "UniWaiting",
    "UniWorker",
    "UniWorkerDefinition",
    "SerializersRegistry",
    "serializer_registry",
    "compressor_registry",
    "ConnectionObj",
    "ConnectionRC",
    "ConnectionManager",
    "ConnectionPool",
    "connection_pool",
)



import json
from typing import Optional, Tuple, Callable, Any, Dict, List

from kafka import KafkaProducer, KafkaConsumer  # type: ignore
from kafka.consumer.fetcher import ConsumerRecord  # type: ignore

from unipipeline.modules.uni_broker import UniBroker, UniBrokerMessageManager
from unipipeline.modules.uni_broker_definition import UniBrokerDefinition
from unipipeline.modules.uni_message_meta import UniMessageMeta


class KafkaUniBrokerMessageManager(UniBrokerMessageManager):
    def reject(self) -> None:
        pass

    def ack(self) -> None:
        pass


class UniKafkaBroker(UniBroker):
    def __init__(self, definition: UniBrokerDefinition[bytes]) -> None:
        super().__init__(definition)

        self._bootstrap_servers = self.get_boostrap_servers()

        self._consumer: Optional[KafkaConsumer] = None
        self._producer: Optional[KafkaProducer] = None

        self._security_conf: Dict[str, Any] = self.get_security_conf()

    def get_boostrap_servers(self) -> List[str]:
        raise NotImplementedError(f'method get_boostrap_server must be implemented for {type(self).__name__}')

    def get_security_conf(self) -> Dict[str, Any]:
        raise NotImplementedError(f'method get_security_conf must be implemented for {type(self).__name__}')

    def connect(self) -> None:
        self._producer = KafkaProducer(
            bootstrap_servers=self._bootstrap_servers,
            api_version=self.definition.kafka_definition.api_version,
            **self._security_conf,
        )

    def close(self) -> None:
        if self._consumer is not None:
            self.stop_consuming()
        if self._producer is not None:
            self.stop_producing()

    def stop_producing(self) -> None:
        assert self._producer is not None
        self._producer.close()
        self._producer = None

    def stop_consuming(self) -> None:
        assert self._consumer is not None
        self._consumer.close()
        self._consumer = None

    def serialize_body(self, meta: UniMessageMeta) -> Tuple[bytes, bytes]:
        meta_dumps = self.definition.message_codec.dumps(meta.dict())
        return str(meta.id).encode('utf8'), bytes(meta_dumps, encoding='utf8')

    def parse_body(self, msg: ConsumerRecord) -> UniMessageMeta:
        if msg.value.get('parent', None) is not None:
            return UniMessageMeta(**msg.value)
        return UniMessageMeta.create_new(data=msg.value)

    def consume(
        self,
        topic: str,
        processor: Callable[[UniMessageMeta, UniBrokerMessageManager], None],
        consumer_tag: str,
        worker_name: str,
        prefetch: int = 1
    ) -> None:

        self._consumer = KafkaConsumer(
            topic,
            api_version=self.definition.kafka_definition.api_version,
            bootstrap_servers=self._bootstrap_servers,
            value_deserializer=lambda m: json.loads(m.decode('utf-8')),
            group_id=worker_name,
        )

        for msg in self._consumer:
            meta = self.parse_body(msg)
            manager = KafkaUniBrokerMessageManager()
            processor(meta, manager)

    def publish(self, topic: str, meta: UniMessageMeta) -> None:
        self.connect()
        key, value = self.serialize_body(meta)
        assert self._producer is not None
        self._producer.send(topic=topic, value=value, key=key)
        self._producer.flush()



from unipipeline import UniWorker

from example.messages.inetermediate_message import InetermediateMessage


class IntermediateSecondWorker(UniWorker):
    def handle_message(self, message: InetermediateMessage) -> None:
        raise NotImplementedError('method handle_message must be specified for class "IntermediateSecondWorker"')



import logging
from typing import Generic, Type, Any, TypeVar, Optional, Dict, Union
from uuid import uuid4

from unipipeline.modules.uni_broker import UniBrokerMessageManager
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_message_meta import UniMessageMeta, UniMessageMetaErrTopic
from unipipeline.modules.uni_worker_definition import UniWorkerDefinition

TMessage = TypeVar('TMessage', bound=UniMessage)
logger = logging.getLogger(__name__)


class UniPayloadParsingError(Exception):
    def __init__(self, exception: Exception):
        self.parent_exception = exception


class UniWorker(Generic[TMessage]):
    def __init__(
        self,
        definition: UniWorkerDefinition,
        mediator: Any
    ) -> None:
        from unipipeline.modules.uni_mediator import UniMediator
        self._uni_moved = False
        self._uni_payload_cache: Optional[TMessage] = None
        self._uni_current_meta: Optional[UniMessageMeta] = None
        self._uni_current_manager: Optional[UniBrokerMessageManager] = None
        self._uni_definition = definition
        self._uni_mediator: UniMediator = mediator
        self._uni_message_type: Type[TMessage] = self._uni_definition.input_message.type.import_class(UniMessage)  # type: ignore
        self._uni_consumer_tag: str = f'{self._uni_definition.name}__{uuid4()}'
        self._uni_worker_instances_for_sending: Dict[Type[UniWorker], UniWorker] = dict()

    @property
    def message_type(self) -> Type[TMessage]:
        return self._uni_message_type

    def consume(self) -> None:
        self._uni_mediator.wait_related_brokers(self._uni_definition.name)
        main_broker = self._uni_mediator.get_connected_broker_instance(self._uni_definition.broker.name)

        self._uni_definition.wait_everything()
        logger.info("worker %s start consuming", self._uni_definition.name)
        main_broker.consume(
            topic=self._uni_definition.topic,
            processor=self.process_message,
            consumer_tag=self._uni_consumer_tag,
            prefetch=self._uni_definition.prefetch,
            worker_name=self._uni_definition.name,
        )

    @property
    def meta(self) -> UniMessageMeta:
        assert self._uni_current_meta is not None
        return self._uni_current_meta

    @property
    def manager(self) -> UniBrokerMessageManager:
        assert self._uni_current_manager is not None
        return self._uni_current_manager

    @property
    def payload(self) -> TMessage:
        if self._uni_payload_cache is None:
            try:
                self._uni_payload_cache = self._uni_message_type(**self.meta.payload)
            except Exception as e:
                raise UniPayloadParsingError(e)
        return self._uni_payload_cache

    def send(self, payload: Union[Dict[str, Any], TMessage], meta: Optional[UniMessageMeta] = None) -> None:
        if isinstance(payload, self._uni_message_type):
            payload_data = payload.dict()
        elif isinstance(payload, dict):
            payload_data = self._uni_message_type(**payload).dict()
        else:
            raise TypeError(f'data has invalid type.{type(payload).__name__} was given')
        meta = meta if meta is not None else UniMessageMeta.create_new(payload_data)
        self._uni_mediator.get_connected_broker_instance(self._uni_definition.broker.name).publish(self._uni_definition.topic, meta)
        logger.info("worker %s sent message %s to %s topic", self._uni_definition.name, meta, self._uni_definition.topic)

    def send_to_worker(self, worker_type: Type['UniWorker[TMessage]'], data: Any) -> None:
        if worker_type not in self._uni_worker_instances_for_sending:
            if not issubclass(worker_type, UniWorker):
                raise ValueError(f'worker_type {worker_type.__name__} is not subclass of UniWorker')
            w_def = self._uni_mediator.get_worker_definition_by_type(worker_type, UniWorker)
            if w_def.name not in self._uni_definition.output_workers:
                raise ValueError(f'worker {w_def.name} is not defined in workers->{self._uni_definition.name}->output_workers')
            w = worker_type(w_def, self._uni_mediator)
            self._uni_worker_instances_for_sending[worker_type] = w

        assert self._uni_current_meta is not None
        meta = self._uni_current_meta.create_child(data)
        self._uni_worker_instances_for_sending[worker_type].send(data, meta=meta)

    def process_message(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        logger.debug("worker %s message %s received", self._uni_definition.name, meta)
        self._uni_moved = False
        self._uni_current_meta = meta
        self._uni_current_manager = manager
        self._uni_payload_cache = None

        unsupported_err_topic = False
        if not meta.has_error:
            try:
                self.handle_message(self.payload)
            except Exception as e:
                logger.error(e)
                self.move_to_error_topic(UniMessageMetaErrTopic.HANDLE_MESSAGE_ERR, e)
        else:
            try:
                assert meta.error is not None  # for mypy needs
                if meta.error.error_topic is UniMessageMetaErrTopic.HANDLE_MESSAGE_ERR:
                    self.handle_error_message_handling(self.payload)
                elif meta.error.error_topic is UniMessageMetaErrTopic.MESSAGE_PAYLOAD_ERR:
                    self.handle_error_message_payload(self.meta, self.manager)
                elif meta.error.error_topic is UniMessageMetaErrTopic.ERROR_HANDLING_ERR:
                    self.handle_error_handling(self.meta, self.manager)
                else:
                    unsupported_err_topic = True
            except Exception as e:
                logger.error(e)
                self.move_to_error_topic(UniMessageMetaErrTopic.ERROR_HANDLING_ERR, e)

        if unsupported_err_topic:
            assert meta.error is not None  # for mypy needs
            err = NotImplementedError(f'{meta.error.error_topic} is not implemented in process_message')
            logger.error(err)
            self.move_to_error_topic(UniMessageMetaErrTopic.SYSTEM_ERR, err)

        if not self._uni_moved and self._uni_definition.ack_after_success:
            manager.ack()

        logger.debug("worker message %s processed", meta)
        self._uni_moved = False
        self._uni_current_meta = None
        self._uni_current_manager = None
        self._uni_payload_cache = None

    def handle_message(self, message: TMessage) -> None:
        raise NotImplementedError(f'method handle_message not implemented for {type(self).__name__}')

    def move_to_error_topic(self, err_topic: UniMessageMetaErrTopic, err: Exception) -> None:
        self._uni_moved = True
        meta = self.meta.create_error_child(err_topic, err)
        self._uni_mediator.get_connected_broker_instance(self._uni_definition.broker.name).publish(f'{self._uni_definition.topic}__{err_topic.value}', meta)
        self.manager.ack()

    def handle_error_message_handling(self, message: TMessage) -> None:
        pass

    def handle_error_message_payload(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        pass

    def handle_error_handling(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        pass

    def handle_uni_error(self, meta: UniMessageMeta, manager: UniBrokerMessageManager) -> None:
        pass



import logging
from collections import deque
from typing import Callable, Dict, TypeVar, Tuple, Optional, Deque

from unipipeline import UniBroker, UniMessageMeta, UniBrokerMessageManager

logger = logging.getLogger(__name__)


class UniMemoryBrokerMessageManager(UniBrokerMessageManager):
    def __init__(self, ql: 'QL', msg_id: int) -> None:
        self._msg_id = msg_id
        self._ql = ql

    def reject(self) -> None:
        self._ql.move_back_from_reserved(self._msg_id)

    def ack(self) -> None:
        self._ql.mark_as_processed(self._msg_id)


TItem = TypeVar('TItem')


TConsumer = Callable[[UniMessageMeta, UniBrokerMessageManager], None]


class QL:
    def __init__(self, name: str) -> None:
        self._name = name
        self._waiting_for_process: Deque[Tuple[int, UniMessageMeta]] = deque()
        self._in_process: Optional[Tuple[int, UniMessageMeta]] = None

        self._msg_counter: int = 0
        self._lst_counter: int = 0
        self._listeners: Dict[int, Tuple[TConsumer, int]] = dict()

    def add_msg(self, msg: UniMessageMeta) -> int:
        self._msg_counter += 1
        msg_id = self._msg_counter
        self._waiting_for_process.append((msg_id, msg))
        logger.debug('ql(%s).add_msg msg_id=%s', self._name, msg_id)
        return msg_id

    def move_back_from_reserved(self, msg_id: int) -> None:
        logger.debug('ql(%s).move_back_from_reserved', self._name)
        if self._in_process is None:
            return

        (msg_id_, meta) = self._in_process
        if msg_id != msg_id_:
            return

        self._waiting_for_process.appendleft((msg_id, meta))

    def reserve_next(self) -> Tuple[int, UniMessageMeta]:
        if self._in_process is not None:
            return self._in_process

        item = self._waiting_for_process.popleft()
        self._in_process = item
        logger.debug('ql(%s).reserve_next msg_id=%s', self._name, item[0])
        return item

    def mark_as_processed(self, msg_id: int) -> None:
        logger.debug('ql(%s).mark_as_processed msg_id=%s', self._name, msg_id)
        if self._in_process is None:
            return

        (msg_id_, meta) = self._in_process
        if msg_id != msg_id_:
            return

        self._in_process = None

    def add_listener(self, listener: TConsumer, prefetch: int) -> int:
        lsg_id = self._lst_counter
        self._lst_counter += 1
        self._listeners[lsg_id] = (listener, prefetch)
        return lsg_id

    def rm_listener(self, lst_id: int) -> None:
        if lst_id not in self._listeners:
            return
        self._listeners.pop(lst_id)

    def messages_to_process_count(self) -> int:
        return len(self._waiting_for_process) + (0 if self._in_process is None else 1)

    def has_messages_to_process(self) -> bool:
        return self.messages_to_process_count() > 0

    def process_all(self) -> None:
        logger.debug('ql(%s).process_all len_listeners=%s :: messages=%s', self._name, len(self._listeners), self.messages_to_process_count())
        if len(self._listeners) == 0:
            return

        while self.has_messages_to_process():
            for lst_id in self._listeners.keys():
                if not self.has_messages_to_process():
                    break

                if lst_id not in self._listeners:
                    continue

                (lst, prefetch) = self._listeners[lst_id]

                for i in range(prefetch):
                    if not self.has_messages_to_process():
                        break

                    (msg_id, meta) = self.reserve_next()
                    manager = UniMemoryBrokerMessageManager(self, msg_id)

                    logger.info('ql(%s).process_all :: lsg_id=%s :: i=%s :: msg_id=%s :: %s', self._name, lst_id, i, msg_id, meta)
                    lst(meta, manager)
                    logger.debug('ql(%s).process_all len_listeners=%s :: messages=%s', self._name, len(self._listeners), self.messages_to_process_count())


class UniMemoryBroker(UniBroker):
    _queues_by_topic: Dict[str, QL] = dict()

    def connect(self) -> None:
        pass

    def close(self) -> None:
        pass

    def _get_ql(self, topic: str) -> QL:
        name = topic
        topic = f'{self.definition.name}:@:@:{topic}'
        if topic not in UniMemoryBroker._queues_by_topic:
            UniMemoryBroker._queues_by_topic[topic] = QL(f'{self.definition.name}->{name}')
        return UniMemoryBroker._queues_by_topic[topic]

    def consume(self, topic: str, processor: TConsumer, consumer_tag: str, worker_name: str, prefetch: int = 1) -> None:
        ql = self._get_ql(topic)
        ql.add_listener(processor, prefetch)
        ql.process_all()

    def publish(self, topic: str, meta: UniMessageMeta) -> None:
        ql = self._get_ql(topic)
        ql.add_msg(meta)
        ql.process_all()



from functools import reduce
from operator import or_
from typing import Dict, Any, Set, Iterator, Tuple


class ParseDefinitionError(Exception):
    pass


def parse_definition(conf_name: str, definitions: Dict[str, Any], defaults: Dict[str, Any], required: Set[str], sys_names: Set[str] = None) -> Iterator[Tuple[str, Dict[str, Any]]]:
    if sys_names is None:
        sys_names = set()

    if not isinstance(definitions, dict):
        raise ParseDefinitionError(f'definition of {conf_name} has invalid type. must be dict')

    max_set_of_keys = reduce(or_, [set(defaults.keys()), required])
    defaults = dict(defaults)
    defaults.update(definitions.get("__default__", dict()))

    for name, raw_definition in definitions.items():
        name = str(name)
        if name.startswith("_"):
            if name == "__default__":
                continue
            elif name in sys_names:
                pass
            else:
                raise ValueError(f"key '{name}' is not acceptable")

        if not isinstance(raw_definition, dict):
            raise ParseDefinitionError(f'definition of {conf_name}->{name} has invalid type. must be dict')

        definition = dict(defaults)
        definition.update(raw_definition)

        definition_keys = set(definition.keys())

        for k, v in definition.items():
            vd = defaults.get(k, None)
            if vd is not None and type(vd) != type(v):
                raise ParseDefinitionError(f'definition of {conf_name}->{name} has invalid key "{k}" type')

        if max_set_of_keys != definition_keys:
            required_props = max_set_of_keys.difference(definition_keys)
            invalid_props = definition_keys.difference(max_set_of_keys)
            if len(invalid_props) > 0:
                raise ParseDefinitionError(f'definition of {conf_name}->{name} has invalid extra props: {", ".join(invalid_props)}')
            if len(required_props) > 0:
                raise ParseDefinitionError(f'definition of {conf_name}->{name} has no required props: {", ".join(required_props)}')
            raise ParseDefinitionError(f'definition of {conf_name}->{name} has invalid props: {max_set_of_keys}!={definition_keys}')

        definition["name"] = name

        yield name, definition



from typing import Optional, Any, Dict, Tuple, TypeVar, Generic

from pydantic import BaseModel, validator

from unipipeline.modules.uni_module_definition import UniModuleDefinition
from unipipeline.utils.serializer_registry import serializer_registry, compressor_registry


TContent = TypeVar('TContent')


class UniMessageCodec(BaseModel, Generic[TContent]):
    compression: Optional[str]
    content_type: str

    @validator("content_type")
    def content_type_must_be_supported(cls, v: str) -> str:
        serializer_registry.assert_supports(v)
        return v

    @validator("compression")
    def compression_must_be_supported(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return None
        compressor_registry.assert_supports(v)
        return v

    def decompress(self, data: bytes) -> bytes:
        if self.compression is not None:
            return compressor_registry.loads(data, self.compression)
        return data

    def loads(self, data: TContent) -> Dict[str, Any]:
        return serializer_registry.loads(data, self.content_type)  # type: ignore

    def compress(self, data: TContent) -> TContent:
        if self.compression is not None:
            data_bytes: bytes
            if isinstance(data, str):
                data_bytes = bytes(data, encoding='utf-8')
            elif isinstance(data, bytes):
                data_bytes = data
            return compressor_registry.dumps(data_bytes, self.compression).decode("utf-8")  # type: ignore
        return data

    def dumps(self, data: Dict[str, Any]) -> str:
        return serializer_registry.dumps(data, self.content_type)


class UniBrokerKafkaPropsDefinition(BaseModel):
    api_version: Tuple[int, int]


class UniBrokerRMQPropsDefinition(BaseModel):
    exchange_name: str
    heartbeat: int
    blocked_connection_timeout: int
    socket_timeout: int
    stack_timeout: int
    exchange_type: str


class UniBrokerDefinition(BaseModel, Generic[TContent]):
    name: str
    type: UniModuleDefinition
    retry_max_count: int
    retry_delay_s: int
    passive: bool
    durable: bool
    auto_delete: bool
    is_persistent: bool
    message_codec: UniMessageCodec[TContent]
    kafka_definition: UniBrokerKafkaPropsDefinition
    rmq_definition: UniBrokerRMQPropsDefinition



import os.path
from importlib import import_module, invalidate_caches
from time import sleep
from typing import NamedTuple, Generic, Type, TypeVar, Any

from unipipeline.utils.template import template

T = TypeVar('T')

uni_broker_template = '''
from typing import Callable

from unipipeline import UniBroker, UniMessageMeta, UniBrokerMessageManager


class {{name}}(UniBroker):
    def connect(self) -> None:
        raise NotImplementedError('method connect must be specified for class "{{name}}"')

    def close(self) -> None:
        raise NotImplementedError('method close must be specified for class "{{name}}"')

    def consume(self, topic: str, processor: Callable[[UniMessageMeta, UniBrokerMessageManager], None], consumer_tag: str, worker_name: str, prefetch: int = 1) -> None:
        raise NotImplementedError('method consume must be specified for class "{{name}}"')

    def publish(self, topic: str, meta: UniMessageMeta) -> None:
        raise NotImplementedError('method publish must be specified for class "{{name}}"')

'''

uni_message_template = '''from unipipeline import UniMessage


class {{name}}(UniMessage):
    pass

'''


uni_worker_template = '''from unipipeline import UniWorker

from {{data.input_message.type.module}} import {{data.input_message.type.class_name}}


class {{name}}(UniWorker):
    def handle_message(self, message: {{data.input_message.type.class_name}}) -> None:
        raise NotImplementedError('method handle_message must be specified for class "{{name}}"')

'''


uni_waiting_template = '''from unipipeline import UniWaiting


class {{name}}(UniWaiting):
    def try_to_connect(self) -> None:
        raise NotImplementedError('method try_to_connect must be specified for class "{{name}}"')

'''

tpl_map = {
    "UniBroker": uni_broker_template,
    "UniMessage": uni_message_template,
    "UniWorker": uni_worker_template,
    "UniWaiting": uni_waiting_template,
}


class UniModuleDefinition(NamedTuple, Generic[T]):
    module: str
    class_name: str

    def import_class(self, class_type: Type[T], auto_create: bool = False, create_template_params: Any = None) -> Type[T]:
        try:
            mdl = import_module(self.module)
        except ModuleNotFoundError:
            if auto_create:
                hierarchy = self.module.split('.')
                path = f'{os.path.join("./", *hierarchy)}.py'
                path_dir = os.path.dirname(path)
                path_init = os.path.join(path_dir, "__init__.py")
                os.makedirs(os.path.dirname(path), exist_ok=True)
                if not os.path.isfile(path_init):
                    with open(path_init, "wt+") as fi:
                        fi.write("")
                with open(path, 'wt') as fm:
                    fm.writelines(template(tpl_map[class_type.__name__], {
                        "data": create_template_params,
                        "name": self.class_name,
                    }))
                for i in range(10):  # because fs has cache time
                    try:
                        mdl = import_module(self.module)
                        break
                    except ModuleNotFoundError:
                        invalidate_caches()
                        sleep(i)
            else:
                raise
        tp = getattr(mdl, self.class_name)
        if not issubclass(tp, class_type):
            ValueError(f'class {self.class_name} is not subclass of {class_type.__name__}')
        return tp



import uuid
from datetime import datetime
from enum import Enum
from typing import Dict, Any, Optional, Type
from uuid import UUID

from pydantic import BaseModel


class UniMessageMetaErrTopic(Enum):
    SYSTEM_ERR = 'error_system'
    MESSAGE_PAYLOAD_ERR = 'error_message_payload'
    HANDLE_MESSAGE_ERR = 'error_message_handling'
    ERROR_HANDLING_ERR = 'error_handling'


class UniMessageMetaErr(BaseModel):
    error_topic: UniMessageMetaErrTopic
    error_type: str
    error_message: str
    retry_times: int


class UniMessageMeta(BaseModel):
    id: UUID
    date_created: datetime
    payload: Dict[str, Any]

    parent: Optional[Dict[str, Any]]
    error: Optional[UniMessageMetaErr]

    @property
    def has_error(self) -> bool:
        return self.error is not None

    @staticmethod
    def create_new(data: Dict[str, Any], error: Optional[UniMessageMetaErr] = None) -> 'UniMessageMeta':
        return UniMessageMeta(
            id=uuid.uuid4(),
            date_created=datetime.now(),
            payload=data,
            parent=None,
            error=error,
        )

    def create_child(self, payload: Dict[str, Any]) -> 'UniMessageMeta':
        return UniMessageMeta(
            id=uuid.uuid4(),
            date_created=datetime.now(),
            payload=payload,
            parent=self.dict(),
            error=None,
        )

    def create_error_child(self, error_topic: UniMessageMetaErrTopic, error: Exception) -> 'UniMessageMeta':
        return UniMessageMeta(
            id=self.id,
            date_created=self.date_created,
            payload=self.payload,
            parent=self.dict(),
            error=UniMessageMetaErr(
                error_topic=error_topic,
                error_type=type(error).__name__,
                error_message=str(error),
                retry_times=self.error.retry_times + 1 if self.error is not None else 0
            ),
        )



import logging
from time import sleep

from unipipeline.modules.uni_broker import UniBroker
from unipipeline.modules.uni_config import UniConfig, ConfigError
from unipipeline.modules.uni_cron_job import UniCronJob
from unipipeline.modules.uni_mediator import UniMediator
from unipipeline.modules.uni_message import UniMessage
from unipipeline.modules.uni_wating import UniWaiting
from unipipeline.modules.uni_worker import UniWorker
from unipipeline.utils.parse_definition import ParseDefinitionError

logger = logging.getLogger(__name__)


class Uni:
    def __init__(self, config_file_path: str) -> None:
        self._config = UniConfig(config_file_path)
        self._mediator = UniMediator(self._config)

    def check_load_all(self, create: bool = False) -> None:
        try:
            for b in self._config.brokers.values():
                b.type.import_class(UniBroker, create, create_template_params=b)

            for m in self._config.messages.values():
                m.type.import_class(UniMessage, create, create_template_params=m)

            for worker in self._config.workers.values():
                worker.type.import_class(UniWorker, create, create_template_params=worker)

            for waiting in self._config.waitings.values():
                waiting.type.import_class(UniWaiting, create, create_template_params=waiting)
        except (ParseDefinitionError, ConfigError) as e:
            print(f"ERROR: {e}")
            exit(1)

    def start_cron(self) -> None:
        cron_jobs = [UniCronJob.new(i, task, self.get_worker(task.worker.name)) for i, task in enumerate(self._config.cron_tasks.values())]

        if len(cron_jobs) == 0:
            return

        logger.info(f'cron jobs defined: {", ".join(cj.name for cj in cron_jobs)}')

        while True:
            delay, tasks = UniCronJob.search_next_tasks(cron_jobs)

            if delay is None:
                return

            logger.info("sleep %s seconds before running the tasks: %s", delay, [cj.name for cj in tasks])

            if delay > 0:
                sleep(delay)

            logger.info("run the tasks: %s", [cj.name for cj in tasks])

            for cj in tasks:
                cj.send()

            sleep(1.1)  # delay for correct next iteration

    def get_worker(self, name: str, singleton: bool = True) -> UniWorker[UniMessage]:
        return self._mediator.get_worker(name, singleton)



from unipipeline import UniWorker

from unipipeline.messages.uni_cron_message import UniCronMessage


class MySuperCronWorker(UniWorker):
    def handle_message(self, message: UniCronMessage) -> None:
        print(f"!!!catch task_name={message.task_name}")



import logging
from time import sleep

from pydantic import BaseModel

from unipipeline.modules.uni_module_definition import UniModuleDefinition
from unipipeline.modules.uni_wating import UniWaiting

logger = logging.getLogger(__name__)


class UniWaitingDefinition(BaseModel):
    name: str
    retry_max_count: int
    retry_delay_s: int
    type: UniModuleDefinition

    def wait(self) -> None:
        waiting_type = self.type.import_class(UniWaiting)
        for try_count in range(self.retry_max_count):
            try:
                w = waiting_type()
                w.try_to_connect()
                logger.info('%s is available in inu', waiting_type.__name__)
                return
            except Exception:
                logger.debug('retry wait for %s [%s/%s]', waiting_type.__name__, try_count, self.retry_max_count)
                sleep(self.retry_delay_s)
                continue
        raise ConnectionError('unavailable connection to %s', waiting_type.__name__)



import logging
import os.path

from unipipeline import Uni


logging.basicConfig(
    level=os.environ.get('LOGLEVEL', logging.DEBUG),
    format='%(asctime)s | %(name)s | %(levelname)s | %(message)s'
)

CWD = os.path.dirname(os.path.abspath(__file__))

u = Uni(f"{CWD}/dag.yml")

u.check_load_all(create=True)

u.get_worker("input_worker").send(dict())

u.get_worker("input_worker").consume()

u.get_worker("my_super_cron_worker").consume()

print("it works!")

u.start_cron()



from setuptools import setup, find_packages

from codecs import open
from os import path

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="unipipeline",
    version="1.1.0",
    description="simple way to build the declarative and distributed data pipelines with python. it supports rabbitmq or kafka as a broker",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aliaksandr-master/unipipeline",
    author="Aliaksandr Master",
    author_email="alxe.master@gmail.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=["jinja2", "pyyaml", "kafka-python", "pika", "pydantic"]
)



from typing import NamedTuple

from unipipeline.modules.uni_worker_definition import UniWorkerDefinition


class UniCronTaskDefinition(NamedTuple):
    name: str
    worker: UniWorkerDefinition
    when: str
